package com.bergerkiller.bukkit.nolagg;

import java.io.DataOutputStream;
import java.io.IOException;
import java.lang.ref.WeakReference;
import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;

import net.minecraft.server.EntityPlayer;
import net.minecraft.server.Packet;
import net.minecraft.server.Packet51MapChunk;

@SuppressWarnings({ "rawtypes", "unchecked" })
public class BufferedPacket51MapChunk extends Packet51MapChunk {
	
	public static String getBufferSizeName() {
		int size = 80 * buffers.size();
		if (size < 1000) {
			return size + " kb";
		} else {
			return ((double) size / 1000) + " mb";
		}
	}
	
	private static class Buffer {
		public Buffer(BufferedPacket51MapChunk packet) {
			this.load(packet);
		}
		public boolean tryLoad(BufferedPacket51MapChunk packet) {
			if (this.packet.get() == null) {
				this.load(packet);
				return true;
			} else {
				return false;
			}
		}
		private void load(BufferedPacket51MapChunk packet) {
			this.packet = new WeakReference<BufferedPacket51MapChunk>(packet);
			packet.rawData = this.data;
			packet.buffer = this;
		}
		public WeakReference<BufferedPacket51MapChunk> packet;
		public final byte[] data = new byte[81920];
	}

	private static List<Buffer> buffers = new ArrayList<Buffer>();
	
	private static boolean cleanupRunning;
	private static Thread thread;
	public static void init() {
		cleanupRunning = true;
		thread = new Thread() {
			public void run() {
				int i = 0;
				while (!this.isInterrupted() && cleanupRunning) {
					try {
						synchronized (buffers) {
							if (i++ == 3 || buffers.size() > 400) {
								Iterator<Buffer> iter = buffers.iterator();
								while (iter.hasNext()) {
									if (iter.next().packet.get() == null) {
										iter.remove();
									}
								}
								i = 0;
							}
						}
						Thread.sleep(10000);	
					} catch (InterruptedException ex) {}
				}
			}
		};
		thread.start();
	}
	public static void deinit() {
		cleanupRunning = false;
		thread.interrupt();
		synchronized (buffers) {
			buffers.clear();
		}
	}
	
	private Buffer buffer;
	public BufferedPacket51MapChunk(net.minecraft.server.Chunk chunk) {
		synchronized (buffers) {
			for (int i = 0; i < buffers.size(); i++) {
				if (buffers.get(i).tryLoad(this)) {
					this.load(chunk);
					return;
				}
			}
			buffers.add(new Buffer(this));	
			this.load(chunk);
		}
	}
	private void load(net.minecraft.server.Chunk chunk) {
		System.arraycopy(chunk.b, 0, this.rawData, 0, 32768);
        System.arraycopy(chunk.g.a, 0, this.rawData, 32768, 16384);
        System.arraycopy(chunk.i.a, 0, this.rawData, 49152, 16384);
        System.arraycopy(chunk.h.a, 0, this.rawData, 65536, 16384);
        this.a = chunk.x << 4;
        this.c = chunk.z << 4;
        this.d = this.f = 16;
        this.e = 128;
	}
	
	public void send(EntityPlayer player) {
		player.netServerHandler.sendPacket(this);
	}
	
	public void a(DataOutputStream dataoutputstream) throws IOException {
		super.a(dataoutputstream);
		this.buffer.packet.clear();
		this.buffer = null;
	}
		
	static {
		try {
			Field a = Packet.class.getDeclaredField("a");
			a.setAccessible(true);
			Map map = (Map) a.get(null);
			map.put(BufferedPacket51MapChunk.class, 51);
		} catch (Exception e) {
			NoLagg.log(Level.SEVERE, "Failed to bind to map packet chunk, total failure imminent!");
			e.printStackTrace();
		}
	}
}
